﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Threading;
using Microsoft.Win32;

namespace VideoThumbnailer
{
    public partial class Window1 : Window
    {
        // Inheritable attached dependency property indicates whether processing is going on
        public static bool GetProcessing(DependencyObject obj)
        {
            return (bool)obj.GetValue(ProcessingProperty);
        }
        public static void SetProcessing(DependencyObject obj, bool value)
        {
            obj.SetValue(ProcessingProperty, value);
        }
        public static readonly DependencyProperty ProcessingProperty = DependencyProperty.RegisterAttached(
            "Processing", typeof(bool), typeof(Window1), new FrameworkPropertyMetadata(false, FrameworkPropertyMetadataOptions.Inherits));

        // Dependency property indicates the file name of the current media
        public string FileName
        {
            get { return (string)GetValue(FileNameProperty); }
            set { SetValue(FileNameProperty, value); }
        }
        public static readonly DependencyProperty FileNameProperty =
            DependencyProperty.Register("FileName", typeof(string), typeof(Window1));

        // Member variables
        private MediaPlayer _mediaPlayer = new MediaPlayer();
        private Queue<TimeSpan> _positionsToThumbnail = new Queue<TimeSpan>();
        private DispatcherTimer _watchdogTimer = new DispatcherTimer();
        private uint[] framePixels;
        private uint[] previousFramePixels;

        public Window1()
        {
            // Initialize window and hook up to events of interest
            InitializeComponent();
            AddHandler(Button.ClickEvent, new RoutedEventHandler(HandleOpenFileButtonClick));
            _mediaPlayer.MediaOpened += new EventHandler(HandleMediaPlayerMediaOpened);
            _mediaPlayer.Changed += new EventHandler(HandleMediaPlayerChanged);
            _watchdogTimer.Interval = TimeSpan.FromSeconds(1);
            _watchdogTimer.Tick += new EventHandler(HandleWatchdogTimerTick);
        }

        private void HandleOpenFileButtonClick(object sender, RoutedEventArgs e)
        {
            // Display open file dialog
            var openFileDialog = new OpenFileDialog
            {
                CheckFileExists = true,
                Filter = "Video Files (*.wmv;*.dvr-ms;*.mpeg;*.mpg;*.avi)|*.wmv;*.dvr-ms;*.mpeg;*.mpg;*.avi|All Files (*)|*"
            };

            // Show dialog
            var result = openFileDialog.ShowDialog();

            // If a file was chosen...
            if (result.HasValue && result.Value)
            {
                // Reset state
                var fileName = openFileDialog.FileName;
                SetProcessing(this, true);
                ThumbnailPanel.Children.Clear();
                FileName = fileName;

                // Open media file
                _mediaPlayer.ScrubbingEnabled = true;
                _mediaPlayer.Open(new Uri(fileName));
            }
        }

        private void HandleMediaPlayerMediaOpened(object sender, EventArgs e)
        {
            // Get details about opened file
            var numberFramesToThumbnail = (ThumbnailPanel.Columns * ThumbnailPanel.Rows);
            var totalMilliseconds = _mediaPlayer.NaturalDuration.TimeSpan.TotalMilliseconds;
            framePixels = new uint[_mediaPlayer.NaturalVideoWidth * _mediaPlayer.NaturalVideoHeight];
            previousFramePixels = new uint[framePixels.Length];

            // Enqueue a position for each frame (at the center of each of the N segments)
            for (int i = 0; i < numberFramesToThumbnail; i++)
            {
                _positionsToThumbnail.Enqueue(TimeSpan.FromMilliseconds((((2 * i) + 1) * totalMilliseconds) / (2 * numberFramesToThumbnail)));
            }

            // Capture the first frame as a baseline
            RenderBitmapAndCapturePixels(previousFramePixels);

            // Seek to the first thumbnail position
            SeekToNextThumbnailPosition();
        }

        private void HandleMediaPlayerChanged(object sender, EventArgs e)
        {
            // If still processing the file (i.e., not done)...
            if (GetProcessing(this))
            {
                // Capture the current frame
                CaptureCurrentFrame(false);
            }
        }

        private void SeekToNextThumbnailPosition()
        {
            // If more frames remain to capture...
            if (0 < _positionsToThumbnail.Count)
            {
                // Seek to next position and start watchdog timer
                _mediaPlayer.Position = _positionsToThumbnail.Dequeue();
                _watchdogTimer.Start();
            }
            else
            {
                // Done; close media file and stop processing
                _mediaPlayer.Close();
                framePixels = null;
                previousFramePixels = null;
                SetProcessing(this, false);
            }
        }

        private void CaptureCurrentFrame(bool forceCapture)
        {
            // Capture the current frame as an ImageSource
            var imageSource = RenderBitmapAndCapturePixels(framePixels);

            // If captured pixels are different than the previous frame...
            if (forceCapture || !framePixels.SequenceEqual(previousFramePixels))
            {
                // Stop the watchdog timer
                _watchdogTimer.Stop();

                // Add an Image for the Thumbnail
                ThumbnailPanel.Children.Add(
                    new Image
                    {
                        Source = imageSource,
                        ToolTip = _mediaPlayer.Position,
                        MaxWidth = _mediaPlayer.NaturalVideoWidth,
                        MaxHeight = _mediaPlayer.NaturalVideoHeight,
                        Margin = new Thickness(2)
                    });

                // Swap the pixel buffers (moves current to previous and avoids allocating a new buffer for current)
                var tempPixels = framePixels;
                framePixels = previousFramePixels;
                previousFramePixels = tempPixels;

                // Seek to the next thumbnail position
                SeekToNextThumbnailPosition();
            }
        }

        private void HandleWatchdogTimerTick(object sender, EventArgs e)
        {
            // Stop the watchdog timer
            _watchdogTimer.Stop();

            // Capture the current frame (even if it's not different than the previous)
            CaptureCurrentFrame(true);
        }

        private ImageSource RenderBitmapAndCapturePixels(uint[] pixels)
        {
            // Render the current frame into a bitmap
            var drawingVisual = new DrawingVisual();
            var renderTargetBitmap = new RenderTargetBitmap(_mediaPlayer.NaturalVideoWidth, _mediaPlayer.NaturalVideoHeight, 96, 96, PixelFormats.Default);
            using (var drawingContext = drawingVisual.RenderOpen())
            {
                drawingContext.DrawVideo(_mediaPlayer, new Rect(0, 0, _mediaPlayer.NaturalVideoWidth, _mediaPlayer.NaturalVideoHeight));
            }
            renderTargetBitmap.Render(drawingVisual);

            // Copy the pixels to the specified location
            renderTargetBitmap.CopyPixels(pixels, _mediaPlayer.NaturalVideoWidth * 4, 0);

            // Return the bitmap
            return renderTargetBitmap;
        }
    }
}
